import factory
import factory.django
import random
import string

from kesha.models import Parent, Account, Booking, Entry
from djmoney.money import Money


class ActiveParentFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Parent

    name = factory.Sequence(lambda n: "Active Parent %03d" % n)
    active = True


class PassiveParentFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Parent

    name = factory.Sequence(lambda n: "Passive Parent %03d" % n)
    active = False


class ActiveAccountFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Account

    name = factory.Sequence(lambda n: "Account %03d" % n)
    parent = factory.SubFactory(ActiveParentFactory)
    virtual = False


class PassiveActiveAccountFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Account

    name = factory.Sequence(lambda n: "Account %03d" % n)
    parent = factory.SubFactory(PassiveParentFactory)
    virtual = False


class EntryFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Entry

    account = factory.SubFactory(ActiveAccountFactory)
    text = "".join(random.choices(string.ascii_letters, k=20))
    virtual = False


class BookingFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Booking

    entry_1 = factory.RelatedFactory(
        EntryFactory, factory_related_name="booking", debit=Money("100.0", "EUR")
    )

    class Params:
        good = factory.Trait(
            entry_2=factory.RelatedFactory(
                EntryFactory,
                factory_related_name="booking",
                credit=Money("100.0", "EUR"),
            )
        )



USE_TZ = False
SECRET_KEY = "fake_key_for_testing"
INSTALLED_APPS = [
    "djmoney",
    "kesha",
    "kesha.tests",
]
DATABASES = dict(
    default=dict(
        ENGINE="django.db.backends.sqlite3",
        NAME=":memory:",
    )
)



from decimal import Decimal
from django.core.exceptions import ValidationError
from django.db.utils import IntegrityError
from django.test import TestCase
from kesha.models import Entry, ModelDoneError
from djmoney.money import Money
from kesha.tests.factories import (
    ActiveParentFactory,
    ActiveAccountFactory,
    BookingFactory,
)


class KeshaTestCase(TestCase):
    def setUp(self):
        self.p = ActiveParentFactory()
        self.a = ActiveAccountFactory()
        self.b = BookingFactory(good=True)

    def test_entry(self):
        """Test if an entry can have both filled debit and credit (shouldn't be possible)."""
        e = Entry(
            account=self.a,
            booking=self.b,
            debit=Money(10.00, "EUR"),
            credit=Money(10.00, "EUR"),
            text="Testbuchung",
        )
        self.assertRaises(IntegrityError, e.save)

    def test_booking_done(self):
        """Tests if a booking can be marked as done, but then not unmarked afterwards."""
        self.b.done = True
        self.b.save()
        self.b.done = False
        self.assertRaises(ModelDoneError, self.b.save)

    def test_booking_done_with_unequal_entries(self):
        b = BookingFactory()
        b.done = True
        self.assertRaises(ValidationError, b.save)

    def test_entry_of_done_booking_not_editable(self):
        """Tests if entries of a booking, which is marked as done, can be changed."""
        self.b.done = True
        self.b.save()
        e = self.b.entries.all()[0]
        e.text = e.text + "update"
        self.assertRaises(ModelDoneError, e.save)

    def test_booking_sum(self):
        """
        Creates a good booking (i.e. the sum of the entries debit = sum of entries credit.
        Also virtual entries are not accounted for.
        """
        b = BookingFactory(good=True)
        Entry.objects.create(
            account=ActiveAccountFactory(),
            booking=b,
            credit=Money(100.00, "EUR"),
            text="Testbuchung",
            virtual=True,
        )
        self.assertEquals(b.debit, Decimal("100.00"))
        self.assertEquals(b.credit, Decimal("100.00"))



from decimal import Decimal
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Sum
from django.utils.text import slugify
from django.utils.translation import gettext as _
from djmoney.models.fields import MoneyField

MODEL_DONE_ERROR_MSG = _("Models marked as done can not be edited anymore.")


class ModelDoneError(Exception):
    def __init__(self, msg=MODEL_DONE_ERROR_MSG, *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class CreatedModifiedModel(models.Model):
    updated_at = models.DateTimeField(auto_now=True)
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        abstract = True


class SlugifiedModel(models.Model):
    slug = models.SlugField(editable=False, unique=True)

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if self._state.adding:
            self.slug = slugify(self.name)
        super().save(*args, **kwargs)


class Parent(CreatedModifiedModel, SlugifiedModel):
    name = models.CharField(max_length=255)
    active = models.BooleanField()


class Account(CreatedModifiedModel, SlugifiedModel):
    name = models.CharField(max_length=255)
    parent = models.ForeignKey(
        "Parent",
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        related_name="child_accounts",
    )
    virtual = models.BooleanField(default=False)

    class Meta:
        unique_together = [["name", "parent"]]

    @property
    def active(self):
        return self.parent.active


class Booking(CreatedModifiedModel):
    done = models.BooleanField(default=False)

    __done = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__done = self.done

    @property
    def debit(self):
        return self.get_entry_sum("debit")

    @property
    def credit(self):
        return self.get_entry_sum("credit")

    def get_entry_sum(self, column):
        value = Entry.objects.filter(booking=self, virtual=False).aggregate(
            Sum(column)
        )[f"{column}__sum"]
        return value if value is not None else Decimal(0.0)

    @property
    def entry_sums_match(self):
        return self.get_entry_sum("debit") == self.get_entry_sum("credit")

    def save(self, force_insert=False, force_update=False, *args, **kwargs):
        if self.done and self.done == self.__done:
            raise ModelDoneError()
        elif self.done and self.done != self.__done:
            if not self.entry_sums_match:
                self.done = False
                raise ValidationError(
                    _(f"Entry sums do not match up Debit {self.debit} != {self.credit}")
                )
            else:
                super().save(force_insert, force_update, *args, **kwargs)
            self.__done = self.done
        elif not self.done and self.done == self.__done:
            super().save(force_insert, force_update, *args, **kwargs)
            self.__done = self.done
        elif not self.done and self.done != self.__done:
            raise ModelDoneError()


class Entry(CreatedModifiedModel):
    account = models.ForeignKey(
        "Account", on_delete=models.PROTECT, related_name="entries"
    )
    booking = models.ForeignKey(
        "Booking", on_delete=models.PROTECT, related_name="entries"
    )
    text = models.TextField()
    debit = MoneyField(
        max_digits=14, decimal_places=2, default_currency="EUR", null=True, blank=True
    )
    credit = MoneyField(
        max_digits=14, decimal_places=2, default_currency="EUR", null=True, blank=True
    )
    virtual = models.BooleanField(default=False)

    class Meta:
        constraints = [
            models.CheckConstraint(
                name="%(app_label)s_%(class)s_either_debit_or_credit",
                check=(
                    models.Q(debit__isnull=True, credit__isnull=False)
                    | models.Q(debit__isnull=False, credit__isnull=True)
                ),
            )
        ]

    def save(self, *args, **kwargs):
        if self.done:
            raise ModelDoneError()
        else:
            super().save(*args, **kwargs)

    @property
    def done(self):
        return self.booking.done



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['kesha', 'kesha.migrations', 'kesha.tests']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.2,<4.0', 'django-money>=1.3.1,<2.0.0']

setup_kwargs = {
    'name': 'django-kesha',
    'version': '0.1.0',
    'description': 'Accounting Helper Django App',
    'long_description': None,
    'author': 'Florian Rämisch',
    'author_email': 'olf@subsignal.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
